Tooltip = class();

-- The 'content' argument can be:
--  * a one-line string, in which case an optimally-sized control will be created for it
--  * a control, something derived from Turbine.UI.Control
--  * a function that returns a string or a control, which will be called each time the Tooltip displays

function Tooltip:Constructor(content)
    self.content = content;
end

function Tooltip:MakeBriefTooltip(str)
    -- A brief tooltip.  Create a control to contain the string.
    local control = Turbine.UI.Lotro.TextBox();
    control:SetBackColor(Turbine.UI.Color(0.75, 0, 0, 0));
    control:SetFont(Turbine.UI.Lotro.Font.TrajanPro16);
    control:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
    
    -- Figure out its optimal size using a temporary label.
    local label = Turbine.UI.Label();
    label:SetFont(Turbine.UI.Lotro.Font.TrajanPro16);
    label:SetText(str);
    label:AutoSize();

    control:SetHeight(label:GetHeight() + 5);
    control:SetWidth(label:GetWidth() + 5);
    control:SetText(str);
    return control;
end

function Tooltip:GetContent()
    if (type(self.content) == "string") then
        -- Content is a string; we need to make our own control for it
        return self:MakeBriefTooltip(self.content);
    elseif (type(self.content) == "function") then
        -- Content is a user-supplied function that returns a string or a Turbine.UI.Control
        local result = self.content(self.attachment);
        if (type(result) == "string") then
            return self:MakeBriefTooltip(result);
        else
            return result;
        end
    else
        -- Content is a user-supplied Turbine.UI.Control
        return self.content;
    end
end

-- Note: This should only be called once; changing the attachment to a different control is not implemented.
function Tooltip:Attach(control)
    self.attachment = control;
    control._ShowTooltip = function()
        self:SetVisible(true);
    end
    control._HideTooltip = function()
        self:SetVisible(false);
    end
    AddCallback(control, "MouseHover", control._ShowTooltip);
    AddCallback(control, "MouseLeave", control._HideTooltip);
end

function Tooltip:Detach()
    if (self.attachment) then
        RemoveCallback(self.attachment, "MouseHover", self.attachment._ShowToolTip);
        RemoveCallback(self.attachment, "MouseLeave", self.attachment._HideToolTip);
    end
    self.attachment = nil;
end

function Tooltip:SetVisible(visible)
    if (visible) then
        if (not self.window) then
            local x, y = Turbine.UI.Display.GetMousePosition();
            self.window = Turbine.UI.Window();
            local content = self:GetContent();
            content:SetParent(self.window);
            self.window.content = content;
            self.window:SetSize(content:GetSize());
            self.window:SetZOrder(2147483647); -- in front of everything
            self.window:SetPosition(x, y + 20);
            self.window:SetVisible(true);
        end
    else
        if (self.window) then
            self.window.content:SetParent(nil);
            self.window:SetVisible(false);
            self.window:Close();
            self.window = nil;
        end
    end
end

Thurallor = Thurallor or {};
Thurallor.UI = Thurallor.UI or {};
Thurallor.UI.Tooltip = Tooltip;
