Event = class(Turbine.Object);

function Event:Constructor(tab, settings, desc, charName, expTime, category)
    Turbine.Object.Constructor(self);

    self.tab = tab;
    self.settings = settings;
    if (desc) then
        self.settings.desc = desc;
    end
    if (charName) then
        self.settings.charName = charName;
    end
    if (expTime) then
        self.settings.expTime = expTime;
    end
    if (category) then
        self.settings.category = category;
    end
end

function Event:GetCells()
    return self.cells;
end

function Event:SelectionGained()
    -- Enable cell updates for size change detection
    -- Lotro.TextBoxes apparently don't get SizeChanged events, so we need this kludge
    for _, cell in pairs(self.cells) do
        cell:SetWantsUpdates(true);
    end
end

function Event:SelectionLost()
    -- Disable cell updates, get rid of pulldowns
    for _, cell in pairs(self.cells) do
        cell:SetWantsUpdates(false);
    end
end

function Event:CleanupDropDowns()
    self.tab:CleanupDropDowns();
end

function Event:SetCells(cells)
    if (cells == nil) then
        cells = {};
    end
    self.cells = cells;
    self:SetDescription(self.settings.desc);
    self:SetCharacterName(self.settings.charName);
    self:SetExpirationTime(self.settings.expTime);
    self:SetCategory(self.settings.category);

    -- Event handlers for "Description" cell
    cells.desc:SetReadOnly(false);
    cells.desc.FocusGained = function(ctl)
        self:CleanupDropDowns();
        ctl:SetForeColor(Turbine.UI.Color.White);
    end
    cells.desc.FocusLost = function(ctl)
        self:SetCategory(self.settings.category); -- restore color
        self.settings.desc = ctl:GetText();
        self:SaveSettings();
        self.tab:ReapplySort("desc");
    end

    -- Event handlers for "Character" cell
    cells.charName.FocusGained = function(ctl)
        self:CleanupDropDowns();
        local charNames = self.tab.window:GetCharNames();
        local dropDown = Thurallor.UI.DropDown(charNames, self.settings.charName);
        dropDown:SetParent(ctl);
        dropDown:SetFont(Turbine.UI.Lotro.Font.Verdana14);
        dropDown:SetZOrder(2);
        dropDown:SetPosition(1, 1);
        local width = ctl:GetWidth() - 2;
        dropDown:SetSize(ctl:GetWidth() - 2, ctl:GetHeight() - 2);
        dropDown:SetExpandedWidth(math.max(width, 200));
        dropDown.ItemChanged = function(_, args)
            self:SetCharacterName(args.Text);
            self:SaveSettings();
            self.tab:ReapplySort("charName");
        end
        ctl.dropDown = dropDown;
        self.tab.dropDownCtl = ctl;
    end
    cells.charName.Update = function(ctl)
        if (ctl.dropDown) then
            ctl.dropDown:SetSize(ctl:GetWidth() - 2, ctl:GetHeight() - 2);
        end
    end
    
    -- Event handlers for "Remaining time" cell
    cells.remaining.FocusGained = function(ctl)
        self:CleanupDropDowns();
        local dialog = ExpireTimeDialog(self.tab.window, self, self.settings.expTime);
        dialog.Ok = function(_, args)
            self:SetExpirationTime(args.expTime);
            self:SaveSettings();
            self.tab:ReapplySort("remaining");
        end
    end

    -- Event handlers for "Category" cell
    cells.category.FocusGained = function(ctl)
        self:CleanupDropDowns();
        local catNames = self.tab.window:GetUserCategories();
        local dropDown = Thurallor.UI.DropDown(catNames, self.settings.category);
        dropDown:SetParent(ctl);
        dropDown:SetFont(Turbine.UI.Lotro.Font.Verdana14);
        dropDown:SetZOrder(2);
        dropDown:SetPosition(1, 1);
        local width = ctl:GetWidth() - 2;
        dropDown:SetSize(ctl:GetWidth() - 2, ctl:GetHeight() - 2);
        dropDown:SetExpandedWidth(math.max(width, 200));
        dropDown.ItemChanged = function(_, args)
            self:SetCategory(args.Text);
            self:SaveSettings();
            self.tab:ReapplySort("category");
        end
        ctl.dropDown = dropDown;
        self.tab.dropDownCtl = ctl;
    end
    cells.category.Update = function(ctl)
        if (ctl.dropDown) then
            ctl.dropDown:SetSize(ctl:GetWidth() - 2, ctl:GetHeight() - 2);
        end
    end
end

function Event:SaveSettings(now)
    self.tab:SaveSettings(now);
end

function Event:GetDescription()
    return self.settings.desc;
end

function Event:SetDescription(desc)
    self.settings.desc = desc;
    local cell = self.cells["desc"];
    if (cell) then
        cell:SetText(desc);
    end
end

function Event:GetCharacterName()
    return self.settings.charName;
end

function Event:SetCharacterName(charName)
    self.settings.charName = charName
    local cell = self.cells["charName"];
    if (cell) then
        cell:SetText(charName);
    end
end

function Event:GetExpirationTime()
    return self.settings.expTime;
end

function Event:SetExpirationTime(expTime)
    self.settings.expTime = expTime;
    self.expired = false;
    self:UpdateTimeRemaining();
end

function Event:GetCategory()
    return self.settings.category;
end

function Event:SetCategory(category)
    self.settings.category = category
    local cell = self.cells["category"];
    if (cell) then
        cell:SetText(category);
    end

    -- Set foreground color of all cells based on category
    local color = self.tab.window:GetCategoryColor(category);
    for _, cell in pairs(self.cells) do
        cell:SetForeColor(color);
    end
end

function Event:GetCategory(category)
    return self.settings.category;
end

function Event:UpdateTimeRemaining()
    local cell = self.cells["remaining"];
    if (not cell) then
        return;
    end
    local remaining = self.settings.expTime - Turbine.Engine.GetGameTime();
    if (remaining <= 0) then
        self.expired = true;
        cell:SetText("Now");
        cell.sortValue = 0;
        self.tab:EventExpired(self);
        return;
    end
    
    -- Format the time display string
    remaining = math.ceil(remaining);
    local s = math.floor(remaining % 60);
    local m = math.floor((remaining % 3600) / 60);
    local h = math.floor((remaining % 86400) / 3600);
    local d = math.floor(remaining / 86400);
    if (remaining < 60) then
        text = s .. " seconds";
    elseif (remaining < 10 * 60) then
        text = m .. "m " .. s .. "s";
    elseif (remaining < 60 * 60) then
        text = m .. " minutes";
    elseif (remaining < 24 * 60 * 60) then
        text = h .. "h " .. m .. "m";
    elseif (remaining < 36 * 60 * 60) then
        text = d .. "d " .. h .. "h " .. m .. "m";
    elseif (remaining < 3 * 24 * 60 * 60) then
        text = d .. "d " .. h .. "h";
    else
        text = d .. " days";
    end

    cell:SetText(text);
    cell.sortValue = remaining;
end

-- Gets called once per second by the window's Update() function.
function Event:Tick()
    if (not self.expired) then
        self:UpdateTimeRemaining();
    end
end

function Event:Destroy()
--Puts("Destroying event: " .. self.settings.desc);
end
