Window = class(Turbine.UI.Lotro.Window);

local importPath = getfenv(1)._.Name;
local imagePath = string.gsub(string.gsub(importPath, "%.Window$", ""), "%.", "/") .. "/Images/";

function Window:Constructor(title, plugin)
    Turbine.UI.Lotro.Window.Constructor(self);
    self:SetResizable(true);
    self:SetMinimumSize(200, 150);
    self:SetText(title);
    self.plugin = plugin;

    local b = Thurallor.UI.RoundButton();
    self.newEventButton = b;
    b:SetParent(self);
    b:SetBackground(imagePath .. "plus.tga");
    b:SetHighlightedBackground(imagePath .. "plus_highlight.tga");
    b:SetMouseDownBackground(imagePath .. "plus_click.tga");
    Thurallor.UI.Tooltip("Add new reminder"):Attach(b);
    b.Click = function() self:NewEvent() end;

    b = Thurallor.UI.RoundButton();
    self.settingsButton = b;
    b:SetParent(self);
    b:SetBackground(0x41101833);
    b:SetDisabledBackground(0x41101833);
    b:SetHighlightedBackground(0x41101832);
    b:SetMouseDownBackground(0x41101834);
    Thurallor.UI.Tooltip("Open the options panel"):Attach(b);
    b.Click = function() self:ShowOptionsPanel() end;

    self.MouseClick = function()
        self:Deselect();
    end
    
    self:CreateIcon();

    self.shimmering = Shimmering();
    self.shimmering:SetWantsUpdates(true);
    
    -- Default settings
    local defaults = {};
    defaults.tabs = {};
    defaults.size = { 600, 150 };
    defaults.position = { 100, 100 };
    defaults.characters = {};
    defaults.categories = {
        { name = "Other";     color = { 1, 192/255, 192/255, 192/255 } }; -- Silver
        { name = "Crafting";  color = { 1,       0, 255/255, 127/255 } }; -- Spring Green
        { name = "Crates";    color = { 1,       0,       0, 255/255 } }; -- Blue
        { name = "Instances"; color = { 1, 255/255,  69/255,       0 } }; -- Orange Red
        { name = "Quests";    color = { 1,       0, 206/255, 209/255 } }; -- Dark Turquoise
        { name = "Tasks";     color = { 1, 255/255, 255/255,       0 } }; -- Yellow
    };
    defaults.defaultCategory = "Other";
    defaults.defaultDelay = 24; -- hours
    defaults.icon = { position = { 0.5, 0.5 }; state = "window" };
    defaults.postponeDelay = 4; -- hours
    defaults.pauseDuringCombat = true;
    self.settings = defaults;
    
    -- Default character settings
    self.charDefaults = {
        autoReminders = {
            crafting = { enabled = true; tab = nil; category = "Crafting" };
            crates =   { enabled = true; tab = nil; category = "Crates" };
        }
    }

    -- Register to receive combat state changes
    self.player = Turbine.Gameplay.LocalPlayer.GetInstance();
    AddCallback(self.player, "InCombatChanged", function()
        self:InCombatChanged();
    end);
    self.chatCallback = function(_, args)
        self:ChatReceived(args.ChatType, args.Message);
    end
    
    -- Get recipe info so we don't have to search for it while processing chat messages
    self:DiscoverRecipes();

    local gameTime = Turbine.Engine.GetGameTime();
    self.lastTick = gameTime;
    self.lastSave = gameTime;
end

function Window:CreateIcon()
    local border = Turbine.UI.Control();
    border:SetBackColor(Turbine.UI.Color(1, 0.8, 0.8, 0.8));
    border:SetSize(32, 32);

    local icon = Turbine.UI.Label();
    icon:SetMouseVisible(false);
    icon:SetParent(border);
    icon:SetPosition(2, 2);
    icon:SetSize(28, 28);
    icon:SetMultiline(false);
    icon:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
    icon:SetFont(Turbine.UI.Lotro.Font.TrajanProBold24);
    icon:SetFontStyle(Turbine.UI.FontStyle.Outline);
    icon:SetForeColor(Turbine.UI.Color(1, 0.8, 0.8, 0.8));
    icon:SetBackColor(Turbine.UI.Color(0.5, 0, 0, 0));
    icon:SetOutlineColor(Turbine.UI.Color(0.75, 0, 0, 0));
    icon.defaultText = "R";
    icon:SetText(icon.defaultText);
    self.iconLabel = icon;
    
    border.MouseEnter = function()
        icon:SetBackColor(Turbine.UI.Color(0.5, 0, 0, 1));
        if (icon.shimmering) then
            icon.shimmering:Remove(icon);
            icon.shimmering:SetWantsUpdates(false);
            icon.shimmering:Close();
            icon.shimmering = nil;
        end
    end
    border.MouseLeave = function()
        icon:SetBackColor(Turbine.UI.Color(0.5, 0, 0, 0));
    end

    local iconifier = Thurallor.UI.Iconifier();
    self.iconifier = iconifier;
    iconifier:SetWindow(self);
    iconifier:SetIcon(border);
    Thurallor.UI.Tooltip(function() return self:GetIconTooltip() end):Attach(iconifier);
    self.iconTooltip = title;

    local function saveIconSettings()
        self.settings.icon = { position = { iconifier:GetFractionalPosition() }; state = iconifier:GetIconState() };
        self:SaveSettings();
    end
    AddCallback(iconifier, "IconStateChanged", function(_, state)
        self:IconStateChanged(state);
        saveIconSettings();
    end);
    AddCallback(iconifier, "IconMoved", saveIconSettings);
end

function Window:IconStateChanged(state)
    if (state == "icon") then
    
        -- Window is being minimized
        self.iconLabel:SetText(self.iconLabel.defaultText);
        self.iconTooltip = nil;
        self.shimmering:SetWantsUpdates(false);
        for name, tab in pairs(self.tabsByName) do
            if (tab.table.marquee) then
                tab.table.marquee:SetWantsUpdates(false);
             end
        end
    else -- (state == "window")
    
        -- Window is being restored
        self.shimmering:SetWantsUpdates(true);
        for name, tab in pairs(self.tabsByName) do
            if (tab.table.marquee) then
                tab.table.marquee:SetWantsUpdates(true);
             end
        end
    end
end

function Window:LoadSettings()
--Puts("Loading settings...");
    Turbine.PluginData.Load(Turbine.DataScope.Server, "Reminders", function(loadedData, args)
        if (loadedData) then
            -- Workaround for Turbine localization bug
            loadedData = ImportTable(loadedData);
            DeepTableCopy(loadedData, self.settings);
--Puts("Load complete.");
        else
--Puts("No data.");
        end
    end);

    self:SetSize(unpack(self.settings.size));
    self:SetPosition(unpack(self.settings.position));
    self.iconifier:SetFractionalPosition(unpack(self.settings.icon.position));
    self.iconifier:SetIconState(self.settings.icon.state);

    -- If first login on this character, add character name to list
    local playerName = self.player:GetName();
    if (not self.settings.characters[playerName]) then
        self.settings.characters[playerName] = {};
    end
    self.charSettings = self.settings.characters[playerName];
    self.characters = {}; -- for character dropdown list
    for name, _ in pairs(self.settings.characters) do
        table.insert(self.characters, name);
        -- Handle added character settings
        local charDefaults = DeepTableCopy(self.charDefaults);
        DeepTableCopy(self.settings.characters[name], charDefaults);
        DeepTableCopy(charDefaults, self.settings.characters[name]);
    end
    
    -- If no tabs exist, create a default one
    if (#self.settings.tabs == 0) then
        table.insert(self.settings.tabs, { name = playerName });
    end

    self:GetOptionsPanel():UpdateDisplay();
    
    self:RebuildTabs();
    
    -- Start chat parsing, updates, etc.
    self:InCombatChanged();
end

--function Window:FindMouseClicks(ctl)
--    AddCallback(ctl, "MouseClick", function()
--        Puts("MouseClick found:" .. tostring(ctl));
--    end);
--    local children = ctl:GetControls();
--    Puts(tostring(ctl) .. " has " .. children:GetCount() .. " children:");
--    for c = 1, children:GetCount() do
--        self:FindMouseClicks(children:Get(c));
--    end
--end

function Window:CreateSavedTabs()    
    for t = 1, #self.settings.tabs do
        local settings = self.settings.tabs[t];
        local tab = self:CreateTab(settings);
    end
    self.tabStack:RedistributeTabs();
    self:BringTabToFront(self:GetFrontTab());
end

function Window:BringTabToFront(tabName)
    local tab = self.tabsByName[tabName];
    tab:BringToFront();
    self.shimmering:Remove(tab.tabText);
    self.charSettings.frontTab = tabName;
    self:SaveSettings();
end

function Window:GetFrontTab()
    local tabName = self.charSettings.frontTab;
    if (not tabName) then
        tabName = self.settings.tabs[1].name;
    end
    local tab = self.tabsByName[tabName];
    if (not tab) then
        tabName = self.settings.tabs[1].name;
        tab = self.tabsByName[tabName];
    end
    self.charSettings.frontTab = tabName;
    return tabName;
end

function Window:Closing(args)
    self:SaveSettings(true);
    self:Iconify();
    args.Cancel = true;
end

function Window:Iconify()
    if (not self.iconifier:IsIcon()) then
        self.iconifier:Iconify();
    end
end

function Window:SaveSettings(now)
    if (now) then
--Puts("Saving...");
        self.settings.pluginVersion = self.plugin:GetVersion();
        -- Workaround for Turbine localization bug
        local saveData = ExportTable(self.settings);
        Turbine.PluginData.Save(Turbine.DataScope.Server, "Reminders", saveData, function()
            self:SettingsSaved();
        end);
        self.lastSave = Turbine.Engine.GetGameTime();
    else
--Puts("Deferred save");
        self.saveNeeded = true;
    end
end

function Window:SettingsSaved()
--Puts("Save complete.");
end

function Window:ShowOptionsPanel()
    self:GetOptionsPanel():UpdateDisplay();
    self.optionsPanel:Show();
end

function Window:UpdateOptionsPanel()
    self:GetOptionsPanel():UpdateDisplay();
end

function Window:GetOptionsPanel()
    if (not self.optionsPanel) then
        self.optionsPanel = OptionsPanel(self, self.settings);
    end
    return self.optionsPanel;
end

function Window:Deselect()
    -- Disable the marquee if the user clicks in the border area
    local tab = self.tabsByName[self:GetFrontTab()];
    tab:Deselect();
end

function Window:CreateTab(settings)
    local tab = Tab(self, settings);
    self.tabStack:AddTabCard(tab);
    self.tabsByName[settings.name] = tab;
    AddCallback(tab, "Click", function(_, args)
        if (args.Button == Turbine.UI.MouseButton.Left) then
            self.charSettings.frontTab = settings.name;
            self:SaveSettings();
        end
    end);
    return tab;
end

function Window:GetTab(tabName)
    return self.tabsByName[tabName];
end

function Window:GetCharNames()
    -- Create a sorted list of character names and a lookup table to check for existence
    local names, exists = {}, self.settings.characters;
    for name, _ in pairs(self.settings.characters) do
        table.insert(names, name);
    end
    table.sort(names);
    return names, exists;
end

function Window:GetUserTabs()
    -- Create a sorted list of tab names and a lookup table to check for existence
    local names, exists = {}, {};
    for n, s in ipairs(self.settings.tabs) do
        table.insert(names, s.name);
        exists[s.name] = n;
    end
    table.sort(names);
    return names, exists;
end

function Window:GetNumUserTabs()
    return #self.settings.tabs;
end

function Window:GetUserCategories()
    -- Create a sorted list of category names and a lookup table to check for existence
    local names, exists = {}, {};
    for n, s in ipairs(self.settings.categories) do
        table.insert(names, s.name);
        exists[s.name] = n;
    end
    table.sort(names);
    return names, exists;
end

function Window:GetCategoryColor(category)
    local catNames, catExists = self:GetUserCategories();
    n = catExists[category];
    if (not n) then
        return Turbine.UI.Color(0.75, 0.75, 0.75);
    end
    local color = self.settings.categories[n].color;
    return Turbine.UI.Color(unpack(color));
end

-- Function must accept a single 'event' argument.
-- Returning nil will continue the loop.
-- Returning 'true' aborts the loop and returns the event object.
function Window:ForEachEventDo(func)
    for _, tab in pairs(self.tabsByName) do
        local event = tab:ForEachEventDo(func);
        if (event) then
            return event;
        end
    end
end

function Window:CreateNewTab()
    local num = 1;
    local name = "New tab";
    while self.tabsByName[name] do
        num = num + 1;
        name = "New tab (" .. num .. ")";
    end
    local settings = { name = name };
    table.insert(self.settings.tabs, settings);
    local tab = self:CreateTab(settings);
    self:MakeRoomForNewTab(tab);
    self:BringTabToFront(name);
    self:SaveSettings();
    self:UpdateOptionsPanel();
end

function Window:CloneTab(name)
    local oldTab = self.tabsByName[name];
    local newSettings = {};
    DeepTableCopy(oldTab.settings, newSettings);
    newSettings.tabLeft = 0;
    local nameStem = string.gsub(name, " %([0-9]+%)$", "");
    local num = 1, newName;
    repeat
        num = num + 1;
        newName = nameStem .. " (" .. num .. ")";
    until (not self.tabsByName[newName]);
    newSettings.name = newName;
    table.insert(self.settings.tabs, newSettings);
    local newTab = self:CreateTab(newSettings);
    self:MakeRoomForNewTab(newTab);
    self:BringTabToFront(newName);
    self:SaveSettings();
    self:UpdateOptionsPanel();
end

-- Shift all existing tabs to the right to make room for a newly-created one.
function Window:MakeRoomForNewTab(newTab)
    for _, tab in pairs(self.tabsByName) do
        if (tab ~= newTab) then
            tab:SetTabLeft(tab:GetTabLeft() + 1);
        end
    end
    self.tabStack:RedistributeTabs();
end

function Window:CreateTabStack()
    if (self.tabStack) then
        self.tabStack:SetParent(nil);
    end
    self.tabStack = Thurallor.UI.TabCardStack();
    self.tabStack:SetParent(self);
    self.tabStack:SetPosition(14, 33);
    self.tabRight = 0;
    self.tabsByName = {};
end

function Window:DeleteTab(name)
    for t = 1, #self.settings.tabs do
        if (self.settings.tabs[t].name == name) then
            table.remove(self.settings.tabs, t);
            break;
        end
    end
    self:RebuildTabs();
    self:UpdateOptionsPanel();
end

function Window:RenameTab(oldName, newName)
    for t = 1, #self.settings.tabs do
        if (self.settings.tabs[t].name == oldName) then
            self.settings.tabs[t].name = newName;
            break;
        end
    end
    self:RebuildTabs();
    self:UpdateOptionsPanel();
end

function Window:RebuildTabs()
    self:CreateTabStack();
    self:CreateSavedTabs();
    self:SizeChanged(); -- includes save
end

function Window:SizeChanged(noSave)
    local width, height = self:GetSize();
    self.tabStack:SetSize(width - 28, self:GetHeight() - 55);
    self.tabStack:RedistributeTabs();
    self.newEventButton:SetPosition(30, 0);
    self.settingsButton:SetPosition(width - 66, 0);
    self.settings.size = { width, height };
    self:SaveSettings();
end

function Window:PositionChanged(noSave)
    local left, top = self:GetPosition();
    self.settings.position = { left, top };
    self:SaveSettings();
end    

function Window:Update()
    -- Call Tick() function of each tab once per second.
    -- To do: Make this smarter, to spread out the event updates over multiple frame updates.
    local gameTime = Turbine.Engine.GetGameTime();
    if (gameTime - self.lastTick >= 1) then
        for _, tab in pairs(self.tabStack.tabCards) do
            tab:Tick(self.lastTick, gameTime);
            self.lastTick = gameTime;
        end
    end
    
    -- Save settings no more often than once a minute.
    if (self.saveNeeded and (gameTime - self.lastSave >= 10)) then
        self:SaveSettings(true);
        self.saveNeeded = false;
    end
end

function Window:InCombatChanged(inCombat)
    local inCombat = self.player:IsInCombat();
    if (inCombat and self.settings.pauseDuringCombat) then
        self:Iconify();
        -- Don't process chat messages during combat
        RemoveCallback(Turbine.Chat, "Received", self.chatCallback);
        -- Don't update events or save settings during combat
        self:SetWantsUpdates(false);
    else
        RemoveCallback(Turbine.Chat, "Received", self.chatCallback);
        AddCallback(Turbine.Chat, "Received", self.chatCallback);
        self:SetWantsUpdates(true);
    end
end

function Window:ChatReceived(channel, message)
--Puts("Channel = " .. tostring(channel) .. "; Message = " .. tostring(message));
    if (channel == Turbine.ChatType.Standard) then
        if (self.charSettings.autoReminders.crafting.enabled) then
            local recipeName = string.match(message, "^You have completed the '(.+)' recipe!");
            if (recipeName and self.recipeCooldown[recipeName]) then
                self:AddCraftingReminder(recipeName);
            end
        end
    elseif (channel == Turbine.ChatType.SelfLoot) then
        if (self.charSettings.autoReminders.crates.enabled) then
            if (string.match(message, "^You open the crate and find ")) then
                self:AddCrateReminder();
            end
        end
    end
end

function Window:DiscoverRecipes()
    local attribs = self.player:GetAttributes();
    if (not attribs.GetProfessionInfo) then
        -- Monster players can't do crafting
        return;
    end
    self.recipeCooldown = {};
    self.recipeProfession = {};
    for profName, prof in pairs(Turbine.Gameplay.Profession) do
        local profInfo = attribs:GetProfessionInfo(prof);
        if (profInfo) then
--Puts("Getting " .. profName .. " recipes...");
            for r = 1, profInfo:GetRecipeCount() do
                local recipe = profInfo:GetRecipe(r);
                local cooldown = recipe:GetCooldown();
                if (cooldown > 0) then
                    local name = recipe:GetName();
--local ip = recipe:GetIngredientPack();
--if (ip) then
--    Puts(recipe:GetTier() .. ":" .. recipe:GetName() .. " " .. ip:GetItemInfo():GetName() .. " x" .. ip:GetRequiredQuantity());
--end
                    self.recipeCooldown[name] = cooldown;
                    self.recipeProfession[name] = profName;
                end
            end
        end
    end
end

function Window:NewEvent()
    local frontTab = self.tabsByName[self:GetFrontTab()];
    local expTime = Turbine.Engine.GetGameTime() + self.settings.defaultDelay * 60 * 60;
    local catNames, catExists = self:GetUserCategories();
    local category = self.settings.defaultCategory;
    if (not catExists[category]) then
        category = self.settings.categories[1].name;
    end
    frontTab:CreateEvent("New reminder", self.player:GetName(), expTime, category);
    self:SaveSettings();
end

function Window:AddCraftingReminder(recipeName)
    local tabName = self.charSettings.autoReminders.crafting.tab;
    local tab = self.tabsByName[tabName];
    if (not tab) then
        tab = self.tabsByName[self:GetFrontTab()];
    end
    local profession = self.recipeProfession[recipeName];
    local cooldown = self.recipeCooldown[recipeName];
    local expTime = cooldown + Turbine.Engine.GetGameTime();
    local category = self.charSettings.autoReminders.crafting.category or self.settings.categories[1].name;
    tab:CreateEvent(profession .. ": " .. recipeName, self.player:GetName(), expTime, category);
    self:SaveSettings();
end

function Window:AddCrateReminder()
    local tabName = self.charSettings.autoReminders.crates.tab;
    local tab = self.tabsByName[tabName];
    if (not tab) then
        tab = self.tabsByName[self:GetFrontTab()];
    end
    local cooldown = 22 * 60 * 60;
    local expTime = cooldown + Turbine.Engine.GetGameTime();
    local targetName = "Unknown crate";
    local target = self.player:GetTarget();
    if (target) then
        targetName = target:GetName();
    end
    tab:CreateEvent("Open '" .. targetName .. "'", self.player:GetName(), expTime, self.charSettings.autoReminders.crates.category);
    self:SaveSettings();
end

function Window:AddNotification(desc, controls)
    -- Add shimmer to the items in the window
    for c, ctl in pairs(controls) do
        self.shimmering:Add(ctl);
    end
    
    -- If the window is minimized, add shimmer to the icon
    if (not self:IsVisible()) then
        self.iconTooltip = desc;
        
        -- Put number of notifications superimposed over the icon
        local num = tonumber(self.iconLabel:GetText()) or 0;
        num = num + 1;
        self.iconLabel:SetText(((num <= 9) and num) or "9+")
        self.iconLabel.shimmering = Shimmering();
        self.iconLabel.shimmering:Add(self.iconLabel);
        self.iconLabel.shimmering:SetWantsUpdates(true);
    end
end

function Window:GetIconTooltip()
    return (self.iconTooltip or self:GetText());
end

-- Dims the window and disables it until the specified child window closes.
function Window:WaitForChild(child)
    if (child and self.mask) then
        -- Child changed.  This will probably never happen, but whatever.
        -- Remove previous child.
        self:WaitForChild(nil);
    end
    if (child) then
        self:SetOpacity(0.75);
        self.mask = Turbine.UI.Control();
        self.mask:SetParent(self);
        self.mask:Focus();
        self.mask:SetSize(self:GetSize());
        self.mask:SetZOrder(2147483647);
        self.mask:SetBackColor(Turbine.UI.Color(0.25, 0, 0, 0));
        self.mask:SetBackColorBlendMode(Turbine.UI.BlendMode.Overlay);
        self.mask.child = child;
        self.mask.MouseClick = function(ctl)
            local x, y = Turbine.UI.Display.GetMousePosition();
            local width, height = ctl.child:GetSize();
            local left = math.max(0, x - width / 2);
            local top = math.max(0, y - height/ 2);
            ctl.child:SetPosition(left, top);
        end
        self.mask.callback = function()
            self:WaitForChild(nil);
        end
        AddCallback(child, "Closing", self.mask.callback);
    else
        RemoveCallback(self.mask.child, "Closing", self.mask.callback);
        self:SetOpacity(1);
        self.mask:SetParent(nil);
        self.mask = nil;
    end
end